﻿using System;
using System.ComponentModel;
using System.IO;
using System.Runtime.InteropServices;
using Microsoft.Win32.SafeHandles;

namespace InfiniTec.IO
{
	/// <summary>
	/// Implements a <see cref="Stream"/> object that reads/writes data through a memory mapping.
	/// </summary>
	public class MemoryMappedFileStream : Stream
	{
		private readonly MemoryMappedFileSafeHandle _Handle;
		private readonly long _Size;
		private readonly FileStream _UnderlyingStream;
		private long _Length;
		private long _Position;


		public MemoryMappedFileStream(string name, string file, FileAccess fileAccess)
		{
			if (string.IsNullOrEmpty(file)) throw new ArgumentNullException("file");

			uint protectionFlags;

			switch (fileAccess)
			{
				case FileAccess.Read:
					protectionFlags = (uint) FileProtection.Readonly;
					break;
				case FileAccess.ReadWrite:
					protectionFlags = (uint) FileProtection.ReadWrite;
					break;
				case FileAccess.Write:
					protectionFlags = (uint) FileProtection.ReadWrite;
					break;
				default:
					throw new InvalidEnumArgumentException("fileAccess", (int) fileAccess, typeof (FileAccess));
			}

			_UnderlyingStream = new FileStream(file, FileMode.Open, fileAccess, FileShare.None);
			_Handle = SafeNativeMethods.CreateFileMapping(_UnderlyingStream.SafeFileHandle, IntPtr.Zero,
			                                              protectionFlags, 0, 0, string.IsNullOrEmpty(name) ? null : "Local\\" + name);

			_Length = _Size = _UnderlyingStream.Length;
		}

		private MemoryMappedFileStream(string name, long size)
		{
			_Size = size;

			uint loSize;
			uint hiSize;

			ConvertLong(_Size, out hiSize, out loSize);

			_Handle = SafeNativeMethods.CreateFileMapping(new SafeFileHandle(new IntPtr(-1), false),
			                                              IntPtr.Zero, (uint) FileProtection.ReadWrite,
			                                              hiSize, loSize, string.IsNullOrEmpty(name) ? null : "Local\\" + name);

			if (_Handle.IsInvalid)
			{
				throw new Win32Exception(Marshal.GetLastWin32Error());
			}
		}

		public override bool CanRead { get { return true; } }

		public override bool CanSeek { get { return true; } }

		public override bool CanWrite { get { return true; } }

		public override long Length { get { return _Length; } }

		public override long Position
		{
			get { return _Position; }
			set
			{
				if (value < 0 || value > _Length)
				{
					throw new ArgumentOutOfRangeException("value", value, "The Position must be between 0 and Length");
				}
			}
		}

		/// <summary>
		/// Creates temporary memory mapped file that is backed by the page file.
		/// </summary>
		/// <param name="size">The maximum size of the file.</param>
		/// <returns></returns>
		public static MemoryMappedFileStream CreatePageFileBackedFile(long size)
		{
			return new MemoryMappedFileStream(null, size);
		}

		public override void Flush() {}

		protected override void Dispose(bool disposing)
		{
			base.Dispose(disposing);

			if (!disposing)
			{
				return;
			}
			_Handle.Dispose();

			if (_UnderlyingStream != null)
			{
				_UnderlyingStream.Dispose();
			}
		}

		private static void ConvertLong(long value, out uint hiUint, out uint loUint)
		{
			hiUint = (uint) (value >> 32) & uint.MaxValue;
			loUint = (uint) value & uint.MaxValue;
		}

		public override int Read(byte[] buffer, int offset, int count)
		{
			if (buffer == null) throw new ArgumentNullException("buffer");

			if (offset < 0 || offset > buffer.Length)
			{
				throw new ArgumentOutOfRangeException("offset", "Offset out of range.");
			}
			if (offset + count > buffer.Length)
			{
				throw new ArgumentOutOfRangeException("buffer", "The buffer is not big enough for the requested data.");
			}

			IntPtr ptr = IntPtr.Zero;

			try
			{
				uint loPos;
				uint hiPos;

				ConvertLong(_Position, out hiPos, out loPos);

				if (_Position + count > _Length)
				{
					count = (int) (_Length - _Position);
				}

				if (count < 1)
				{
					return 0;
				}

				ptr = SafeNativeMethods.MapViewOfFile(_Handle, FileMapAccesses.Read, hiPos, loPos, (uint) count);

				if (ptr == IntPtr.Zero)
				{
					throw new Win32Exception(Marshal.GetLastWin32Error());
				}

				Marshal.Copy(ptr, buffer, offset, count);

				_Position += count;

				return count;
			}
			finally
			{
				if (ptr != IntPtr.Zero)
				{
					SafeNativeMethods.UnmapViewOfFile(ptr);
				}
			}
		}

		public override long Seek(long offset, SeekOrigin origin)
		{
			long newPosition;

			switch (origin)
			{
				case SeekOrigin.Begin:
					newPosition = offset;
					break;
				case SeekOrigin.Current:
					newPosition = _Position + offset;
					break;
				case SeekOrigin.End:
					newPosition = _Length + offset;
					break;
				default:
					throw new InvalidEnumArgumentException("origin", (int) origin, typeof (SeekOrigin));
			}

			if (newPosition < 0 || newPosition > Length)
			{
				throw new ArgumentOutOfRangeException("offset", newPosition, "The new position is out of range.");
			}
			_Position = newPosition;

			return newPosition;
		}

		public override void SetLength(long value)
		{
			if (value < 0 || value > _Size)
			{
				throw new ArgumentOutOfRangeException("value", value, string.Format("The length must be between 0 and {0}", value));
			}
		}

		public override void Write(byte[] buffer, int offset, int count)
		{
			if (buffer == null) throw new ArgumentNullException("buffer");

			if (offset < 0 || offset > buffer.Length)
			{
				throw new ArgumentOutOfRangeException("offset", "Offset out of range.");
			}
			if (offset + count > buffer.Length)
			{
				throw new ArgumentOutOfRangeException("buffer", "The specified range exceeds the size of the buffer.");
			}
			if (Position + count > _Size)
			{
				throw new ArgumentOutOfRangeException("count", "Cannot write beyond the size of the file.");
			}

			IntPtr ptr = IntPtr.Zero;

			try
			{
				uint loPos;
				uint hiPos;

				ConvertLong(_Position, out hiPos, out loPos);
				ptr = SafeNativeMethods.MapViewOfFile(_Handle, FileMapAccesses.Write, hiPos, loPos, (uint) count);

				if (ptr == IntPtr.Zero)
				{
					throw new Win32Exception(Marshal.GetLastWin32Error());
				}

				Marshal.Copy(buffer, offset, ptr, count);

				_Position += count;
				if (_Position > _Length)
				{
					_Length = _Position;
				}
			}
			finally
			{
				if (ptr != IntPtr.Zero)
				{
					SafeNativeMethods.UnmapViewOfFile(ptr);
				}
			}
		}
	}
}