using System;
using System.Collections.Generic;
using System.Text;
using System.Security.Principal;
using System.Runtime.InteropServices;
using System.Runtime.CompilerServices;

namespace InfiniTec.Security
{
    /// <summary>
    /// This class provides function for enumeration and manipulation of windows privileges
    /// granted to <see cref="WindowsIdentity"/>.
    /// </summary>
    public class PrivilegeCollection: IEnumerable<string>
    {
        private Dictionary<string, Privilege> _Privileges;
        private WindowsIdentity _Identity;

        /// <summary>
        /// Returns the <see cref="WindowsIdentity"/> associated with this instance.
        /// </summary>
        public WindowsIdentity Identity
        {
            get { return _Identity; }
        }

        /// <summary>
        /// Returns the number of privileges held by the associated <see cref="Identity"/>
        /// </summary>
        public int Count
        {
            get { return _Privileges.Count; }
        }

        /// <summary>
        /// Removes all unnecessary priviliges from the associated <see cref="Identity"/> 
        /// </summary>
        public void ReducePrivilegesToMinimum()
        {
            Privilege changeNotifyPrivilege;

            List<PrivilegeAction> privilegesToRemove;
            changeNotifyPrivilege = null;

            privilegesToRemove = new List<PrivilegeAction>();

            foreach (KeyValuePair<string, Privilege> entry in _Privileges)
            {
                if (entry.Key.Equals(Privilege.ChangeNotify, StringComparison.InvariantCultureIgnoreCase))
                {
                    changeNotifyPrivilege = entry.Value;
                    continue;
                }

                privilegesToRemove.Add(new PrivilegeAction(entry.Value.Luid, PrivilegeState.Remove));
            }
            Privilege.AdjustTokenPrivileges(_Identity, false, privilegesToRemove.ToArray());
            _Privileges.Clear();

            if (changeNotifyPrivilege != null) _Privileges.Add(changeNotifyPrivilege.Name, changeNotifyPrivilege);
        }

        /// <summary>
        /// Creates a new instance of this class for the specified <see cref="Identity"/>
        /// </summary>
        /// <param name="identity">The <see cref="WindowsIdentity"/> to use</param>
        public PrivilegeCollection(WindowsIdentity identity)
        {
            _Privileges = new Dictionary<string, Privilege>();

            _Identity = identity;
            Refresh();
        }

        /// <summary>
        /// Creates a new instance of this class using the current <see cref="WindowsIdentity"/>
        /// </summary>
        public PrivilegeCollection() : this(WindowsIdentity.GetCurrent()) { }

        /// <summary>
        /// Returns the privilege with the specified name.
        /// </summary>
        /// <param name="privilegeName">The name of the privilege</param>
        /// <returns>The <see cref="Privilege"/>, if that privilege is available in the associated <see cref="WindowsIdentity"/></returns>
        /// <exception cref="PrivilegeNotHeldException">If the specified privilege is not held by the <see cref="WindowsIdentity"/></exception>
        public Privilege this[string privilegeName]
        {
            get
            {
                if (!_Privileges.ContainsKey(privilegeName)) throw new PrivilegeNotHeldException("The requested privilege is not held by the specified token.");

                return _Privileges[privilegeName];
            }
        }

        /// <summary>
        /// Reloads the current state of all privileges from the <see cref="WindowsIdentity"/>
        /// </summary>
        public void Refresh()
        {
            uint size;
            Privilege privilege;
            int lastError;
            TOKEN_PRIVILEGES result;

            if (!NativeMethods.GetTokenInformation(_Identity.Token, TokenInformationClass.TokenPrivileges,
                SafeTokenPrivileges.NullHandle, 0, out size))
            {
                lastError = Marshal.GetLastWin32Error();

                if (lastError != 122) // 122 = Databuffer to small; this is the expected value
                    throw Marshal.GetExceptionForHR(lastError);
            }

            using (SafeTokenPrivileges handle = new SafeTokenPrivileges(size))
            {
                if (!NativeMethods.GetTokenInformation(_Identity.Token, TokenInformationClass.TokenPrivileges,
                    handle, size, out size))
                    throw Marshal.GetExceptionForHR(Marshal.GetLastWin32Error());
                result = handle.MarshalNativeToManaged();
            }

            _Privileges.Clear();
            foreach (LUID_AND_ATTRIBUTES entry in result.Privileges)
            {
                privilege = new Privilege(this, _Identity, entry.Luid, (PrivilegeState)entry.Attributes);

                _Privileges.Add(privilege.Name, privilege);
            }

        }

        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return _Privileges.Keys.GetEnumerator();
        }

        #endregion

        #region IEnumerable<string> Members

        /// <summary>
        /// Returns an <see cref="IEnumerator<string>" /> which can be used to enumerate
        /// through the names of associated <see cref="WindowsIdentity"/>
        /// </summary>
        /// <returns>An <see cref="IEnumerator<string>"/> instance.</returns>
        public IEnumerator<string> GetEnumerator()
        {
            return _Privileges.Keys.GetEnumerator();
        }

        #endregion

        internal void RemovePrivilege(string name)
        {
            _Privileges.Remove(name);
        }
    }
}
