﻿using System;
using System.Collections.Generic;
using System.Threading;

namespace InfiniTec.Threading
{
	public static class Dispatcher
	{
		private static Thread _DispatcherThread;
		private static readonly List<DispatcherCookie> _PendingCallbacks = new List<DispatcherCookie>();
		private static readonly object _PendingCallbacksLock = new object();
		private static readonly AutoResetEvent _DispatcherSignal = new AutoResetEvent(false);


		public static DispatcherCookie RegisterCallback(DateTime callbackTime, WaitCallback callback)
		{
			return RegisterCallback(callbackTime, callback, null);
		}

		public static DispatcherCookie RegisterCallback(DateTime callbackTime, WaitCallback callback, object state)
		{
			if (callback == null) throw new ArgumentNullException("callback");
			if (callbackTime < DateTime.Now) throw new ArgumentOutOfRangeException("callbackTime", callbackTime, "The callback time must be in the future.");

			DispatcherCookie cookie = new DispatcherCookie(callbackTime, callback, state);

			lock (_PendingCallbacksLock)
			{
				if (_DispatcherThread == null)
				{
					InitializeDispatcherThread();
				}

				_PendingCallbacks.Add(cookie);
				_PendingCallbacks.Sort();
			}
			_DispatcherSignal.Set();

			return cookie;
		}

		public static void CancelCallback(DispatcherCookie cookie)
		{
			if (cookie == null) throw new ArgumentNullException("cookie");

			lock (_PendingCallbacksLock)
			{
				if (!_PendingCallbacks.Remove(cookie))
				{
					throw new ArgumentException("The cookie is not registered on the dispatcher.");
				}
			}
		}

		private static void InitializeDispatcherThread()
		{
			_DispatcherThread = new Thread(DispatcherThread);
			_DispatcherThread.Name = "infinitec.threading Dispatcher thread";
			_DispatcherThread.IsBackground = true;

			_DispatcherThread.Start();
		}

		private static void DispatcherThread()
		{
			while (true)
			{
				DispatcherCookie cookie = null;
				TimeSpan? delay;

				lock (_PendingCallbacksLock)
				{
					DateTime now = DateTime.Now;
					delay = null;

					if (_PendingCallbacks.Count > 0)
					{
						cookie = _PendingCallbacks[0];
						if (cookie.CallbackTime <= now)
						{
							_PendingCallbacks.RemoveAt(0);
						}
						else
						{
							delay = cookie.CallbackTime - now;
							cookie = null;
						}
					}
				}

				if (delay != null)
				{
					_DispatcherSignal.WaitOne(delay.Value, true);
				}
				else if (cookie == null)
				{
					_DispatcherSignal.WaitOne();
				}
				else
				{
					ThreadPool.QueueUserWorkItem(cookie.Callback, cookie.State);
				}
			}
		}
	}
}
