using System;
using InfiniTec.Diagnostics;
using InfiniTec.Exchange.Eventing.Interop;

namespace InfiniTec.Exchange.Eventing
{
    public abstract class AsynchronousEventSink : EventSink, IExStoreAsyncEvents
    {
        void IExStoreAsyncEvents.OnSave(IExStoreEventInfo pEventInfo, string bstrURLItem, EventFlags lFlags)
        {
            Trace.ActivityId = Guid.NewGuid();
            Trace.Verbose("OnSave called for item {0}.", bstrURLItem);

            ProcessCommitOperation(bstrURLItem, pEventInfo, lFlags, OperationType.Save);
        }

        void IExStoreAsyncEvents.OnDelete(IExStoreEventInfo pEventInfo, string bstrURLItem, EventFlags lFlags)
        {
            Trace.ActivityId = Guid.NewGuid();
            Trace.Verbose("OnDelete called for item {0}.", bstrURLItem);

            ProcessCommitOperation(bstrURLItem, pEventInfo, lFlags, OperationType.Delete);
        }

        internal void ProcessCommitOperation(string url, IExStoreEventInfo eventInfo, EventFlags flags,
                                             OperationType operationType)
        {
            string subject = "(unknown)";

            using (EventInfo info = GetEventInfo((IExStoreDispEventInfo) eventInfo, flags, url))
            {
                try
                {
                    if (!Initialize())
                    {
                        Trace.Verbose("The initialization has failed. Aborting.");
                        return;
                    }

                    var stateBag = info.InternalEventInfo.Data != IntPtr.Zero ? StateBag.DeserializeFromUnmanagedMemory(info.InternalEventInfo.Data)
                                       : new StateBag();

                    subject = GetSubject(info);

                    info.UserState = stateBag.UserState;

                    if (stateBag.IsSilentSaveOperation)
                    {
						Trace.Verbose("Commit phase of a silent save operation. Aborting.");
                        return;
                    }

                    if ((flags & EventFlags.EVT_COPY) == EventFlags.EVT_COPY)
                    {
                        Trace.Verbose("The item {0} is copied here from {1}", info.Url, info.SourceUrl);

                        OnItemCreated(info, CreationMode.Copied);
                    }
                    else if ((flags & EventFlags.EVT_MOVE) == EventFlags.EVT_MOVE && operationType == OperationType.Save)
                    {
                        Trace.Verbose("The item {0} is moved here from {1}", info.Url, info.SourceUrl);
                        OnItemCreated(info, CreationMode.Moved);
                    }
                    else if ((flags & EventFlags.EVT_MOVE) == EventFlags.EVT_MOVE &&
                             operationType == OperationType.Delete)
                    {
                        Trace.Verbose("The item {0} has been moved to the recycle-bin.", info.Url);
                        OnItemDeleted(info, DeletionMode.MovedToRecycleBin);
                    }
                    else if ((flags & EventFlags.EVT_HARDDELETE) == EventFlags.EVT_HARDDELETE)
                    {
                        Trace.Verbose("The item {0} was hard-deleted from the store.", info.Url);
                        OnItemDeleted(info, DeletionMode.Hard);
                    }
                    else if ((flags & EventFlags.EVT_SOFTDELETE) == EventFlags.EVT_SOFTDELETE)
                    {
                        Trace.Verbose("The item {0} has been soft-deleted from the store.",
                                           info.Url);
                        OnItemDeleted(info, DeletionMode.Soft);
                    }
                    else if ((flags & EventFlags.EVT_IS_DELIVERED) == EventFlags.EVT_IS_DELIVERED)
                    {
                        Trace.Verbose("A new item has been delivered to {0}", info.Url);
                        OnItemCreated(info, CreationMode.Delivered);
                    }
                    else if ((flags & EventFlags.EVT_NEW_ITEM) == EventFlags.EVT_NEW_ITEM)
                    {
                        Trace.Verbose("A new item was created at {0}", info.Url, info.SourceUrl);
                        OnItemCreated(info, CreationMode.NewItem);
                    }
                    else if ((flags & EventFlags.EVT_REPLICATED_ITEM) == EventFlags.EVT_REPLICATED_ITEM)
                    {
                        Trace.Verbose("An item has been replicated to {0}", info.Url,
                                           info.SourceUrl);
                        OnItemCreated(info, CreationMode.Replicated);
                    }
                    else
                    {
                        Trace.Verbose("The item {0} has been updated.", info.Url);
                        OnItemUpdated(info);
                    }
                }
                catch (Exception ex)
                {
                    Trace.Error("An unhandled exception occured while processing item at {0} (Subject {1}): {2}",
                                     info.Url, subject, ex);
                    OnErrorOccured(info, ex);
                }
            }
        }

        protected static string GetSubject(EventInfo info)
        {
            string subject;
            try
            {
                subject = (string) info.Item.Fields[WellknownProperties.Item.Subject].Value;
            }
            catch
            {
                subject = "(Unable to retrieve subject)";
            }
            return subject;
        }

        protected virtual void OnItemUpdated(EventInfo info)
        {
        }

        /// <summary>
        /// When called during an asynchronous event, this method is called after the element has been removed and access to 
        /// the item through <see cref="EventInfo.Item"/> is not possible. Otherwise, this method is
        /// called during the commit phase of the event.
        /// </summary>
        /// <param name="info">A <see cref="EventInfo"/> instance containing detailed informations about the item being deleted.</param>
        /// <param name="deletionMode">A value of the <see cref="DeletionMode"/> specifying whether the has been deleted or moved to the recycle bin.</param>
        /// <remarks>
        /// <para>Note to implementors: The Exchange store is blocked during the execution of this method if called during a synchronous event. 
        /// This method must be executed as fast as possible and no blocking calls should be made from this thread. 
        /// Long running operations should be executed on a different thread.</para>
        /// <para>Thread safety: This method is thread safe.</para>
        /// </remarks>
        protected virtual void OnItemDeleted(EventInfo info, DeletionMode deletionMode)
        {
        }

        /// <summary>
        /// When called during an asynchronous event, this method is called after the element has been created and read/write access
        /// is possible to the affected item. Otherwise, this method is called during the commit phase of the event and access to the item
        /// is read-only.
        /// </summary>
        /// <param name="info">A <see cref="EventInfo"/> instance containing detailed informations about the item being created.</param>
        /// <param name="creationMode">A value of the <see cref="CreationMode"/> specifying whether the has been created, copied or moved to this location.</param>
        /// <remarks>
        /// <para>Note to implementors: The Exchange store is blocked during the execution of this method if called during a synchronous event. 
        /// This method must be executed as fast as possible and no blocking calls should be made from this thread. 
        /// Long running operations should be executed on a different thread.</para>
        /// <para>Thread safety: This method is thread safe.</para>
        /// </remarks>
        protected virtual void OnItemCreated(EventInfo info, CreationMode creationMode)
        {
        }
    }
}