using System;
using System.Collections.Generic;
using System.Net;
using System.Runtime.InteropServices;
using ADODB;

namespace InfiniTec.Exchange.Eventing
{
    public class EventRegistration
    {
        public const int DefaultPriority = 0xffff;

        private _Connection _Connection;
        private Record _Record;

        public string Url { get; set; }

        public MatchScope? Scope
        {
            get
            {
                EnsureItemIsOpen();

                var value = (string)_Record.Fields[WellknownProperties.EventRegistration.MatchScope].Value;
                return (string.IsNullOrEmpty(value)) ? (MatchScope?) null : (MatchScope) Enum.Parse(typeof (MatchScope), value, true);
            }
            set
            {
                EnsureItemIsOpen();

                if (value != Scope)
                {
                    _Record.Fields[WellknownProperties.EventRegistration.MatchScope].Value = (value == null) ? string.Empty : value.ToString().ToUpperInvariant();
                }
            }
        }

        public EventMethods EventMethods
        {
            get
            {
                EnsureItemIsOpen();

                var value = (string)_Record.Fields[WellknownProperties.EventRegistration.EventMethod].Value;

                EventMethods result = 0;

                foreach (string entry in value.Split(new[] {';'}, StringSplitOptions.RemoveEmptyEntries))
                {
                    var method = (EventMethods) Enum.Parse(typeof (EventMethods), entry, true);
                    result |= method;
                }
                return result;
            }
            set
            {
                EnsureItemIsOpen();

                if ((value & EventMethods.SynchronousEvents) != EventMethods.None &&
                    (value & EventMethods.AsynchronousEvents) != EventMethods.None)
                {
                    throw new ArgumentException("The specified methods must all be from a singly category.", "value");
                }
                var values = new List<string>();

                if ((value & EventMethods.OnDelete) == EventMethods.OnDelete)
                {
                    values.Add("OnDelete");
                }

                if ((value & EventMethods.OnSave) == EventMethods.OnSave)
                {
                    values.Add("OnSave");
                }
                if ((value & EventMethods.OnSyncDelete) == EventMethods.OnSyncDelete)
                {
                    values.Add("OnSyncDelete");
                }
                if ((value & EventMethods.OnSyncSave) == EventMethods.OnSyncSave)
                {
                    values.Add("OnSyncSave");
                }

                _Record.Fields[WellknownProperties.EventRegistration.EventMethod].Value = string.Join(";", values.ToArray());
            }
        }

        public string Criteria
        {
            get
            {
                EnsureItemIsOpen();

                return (string) _Record.Fields[WellknownProperties.EventRegistration.Criteria].Value;
            }
            set
            {
                EnsureItemIsOpen();

                _Record.Fields[WellknownProperties.EventRegistration.Criteria].Value = value ?? string.Empty;
            }
        }

        public int? Priority
        {
            get
            {
                EnsureItemIsOpen();

                object value = _Record.Fields[WellknownProperties.EventRegistration.Priority].Value;
                return (value != null) ? (int?) value : null;
            }
            set
            {
                EnsureItemIsOpen();

                _Record.Fields[WellknownProperties.EventRegistration.Priority].Value = value.GetValueOrDefault(DefaultPriority);
            }
        }

        public bool? IsEnabled
        {
            get
            {
                EnsureItemIsOpen();

                object value = _Record.Fields[WellknownProperties.EventRegistration.Enabled].Value;
                return (value != null) ? (bool?) value : null;
            }
            set
            {
                EnsureItemIsOpen();

                _Record.Fields[WellknownProperties.EventRegistration.Enabled].Value = value.GetValueOrDefault(true);
            }
        }

        public string SinkClass
        {
            get
            {
                EnsureItemIsOpen();

                return (string)_Record.Fields[WellknownProperties.EventRegistration.SinkClass].Value;
            }
            set
            {
                EnsureItemIsOpen();

                _Record.Fields[WellknownProperties.EventRegistration.SinkClass].Value = value;
            }
        }

        public NetworkCredential Credential { get; set; }

        private static _Connection OpenConnection(string url, NetworkCredential credential)
        {
            var connection = new ConnectionClass
                                 {
                                     Provider = Constants.ExOleDBDataSource,
                                     ConnectionString = url
                                 };

            connection.Open(url, GetUsername(credential), GetPassword(credential), 0);

            return connection;
        }

        public static void Delete(string url, NetworkCredential credential)
        {
            _Connection connection = OpenConnection(url, credential);

            Record record = OpenItem(url, credential, connection);

            record.DeleteRecord(url, false);

            connection.Close();
        }

        public void Delete()
        {
            EnsureItemIsOpen();

            _Record.DeleteRecord(Url, false);
        }

        public void Save()
        {
            if (_Record == null)
            {
                return;
            }

            if (string.IsNullOrEmpty(SinkClass))
            {
                throw new InvalidOperationException("The SinkClass is not set.");
            }

            _Record.Fields[WellknownProperties.Item.ContentClass].Value = Constants.EventItemClass;
            _Record.Fields.Update();

            _Connection.CommitTrans();
        }


        public static EventRegistration Create(string url, NetworkCredential credential)
        {
            var record = new Record();
            _Connection connection = OpenConnection(url, credential);

            connection.BeginTrans();

            record.Open(url, connection, ConnectModeEnum.adModeReadWrite, RecordCreateOptionsEnum.adCreateNonCollection,
                        RecordOpenOptionsEnum.adOpenRecordUnspecified, "", "");

            return new EventRegistration
                       {
                           Url = url,
                           _Record = record,
                           _Connection = connection
                       };
        }

        private static string GetPassword(NetworkCredential credential)
        {
            return (credential != null) ? credential.Password : null;
        }

        private void EnsureItemIsOpen()
        {
            if (_Record != null)
            {
                return;
            }

            _Connection = OpenConnection(Url, Credential);

            if (Url == null)
            {
                throw new InvalidOperationException("The Url parameter must be set before accessing the item.");
            }

            _Record = OpenItem(Url, Credential, _Connection);
        }

        private static Record OpenItem(string url, NetworkCredential credential, _Connection connection)
        {
            if (url == null) throw new ArgumentNullException("url");


            var record = new Record();
            record.Open(url, connection, ConnectModeEnum.adModeReadWrite, RecordCreateOptionsEnum.adFailIfNotExists, RecordOpenOptionsEnum.adOpenSource,
                        GetUsername(credential), GetPassword(credential));

            return record;
        }

        private static string GetUsername(NetworkCredential credential)
        {
            if (credential == null)
            {
                return null;
            }

            return (string.IsNullOrEmpty(credential.Domain)) ? string.Format("{0}\\{1}", credential.Domain, credential.UserName) : credential.UserName;
        }

        public void SetSinkClassFromType<T>()
            where T : EventSink
        {
            SinkClass = Marshal.GenerateProgIdForType(typeof (T));
        }

       
    }
}