﻿using System;
using System.Configuration;
using System.EnterpriseServices;
using System.Text;
using System.Threading;
using ADODB;
using InfiniTec.Diagnostics;
using InfiniTec.Exchange.Eventing.Interop;

namespace InfiniTec.Exchange.Eventing
{
    public abstract class EventSink : ServicedComponent, ICreateRegistration
    {
        private static int _Initialized;
        private static bool _InitializationFailed;

        internal static Trace Trace { get; private set; }

        int ICreateRegistration.Register(IExStoreEventInfo pEventInfo, string bstrURLItem, EventFlags lFlags)
        {
            Trace.ActivityId = Guid.NewGuid();
            Trace.Verbose("Eventsink is registered on {0}", bstrURLItem);
            using (EventInfo info = GetEventInfo((IExStoreDispEventInfo) pEventInfo, lFlags, bstrURLItem))
            {
                try
                {
                    if (!Initialize())
                    {
                        Trace.Verbose("The initialization has failed. Aborting.");
                        return 1;
                    }


                    Trace.Verbose("Calling OnRegisteringEvent");
                    bool result = OnRegisteringEvent(info);

                    if (result)
                    {
                        Trace.Verbose("OnRegistering returned true. Allowing the registration process to complete.");
                        return 0;
                    }
                    Trace.Verbose("OnRegistering returned false. Aborting the registration process.");
                    return 1;
                }
                catch (Exception ex)
                {
                    Trace.Error("The following unhandled occured during the registration process: {0}", ex);
                    OnErrorOccured(info, ex);
                    return 0;
                }
            }
        }

        protected bool Initialize()
        {
            if (_InitializationFailed)
            {
                return false;
            }
            if (Interlocked.Exchange(ref _Initialized, 1) == 1)
            {
                Trace.Verbose("The eventsink has already been initialized.");
                return true;
            }

            Trace.Verbose("Calling OnInitialize");
            try
            {
                OnInitialize();
                Trace.Verbose("Initialization completed");
                return true;
            }
            catch (ConfigurationErrorsException ex)
            {
                _InitializationFailed = true;

                var sb = new StringBuilder();
                sb.AppendFormat("The configuration in file {0} is invalid. The following errors have been found:", ex.Filename);
                sb.AppendLine();
                foreach (ConfigurationErrorsException error in ex.Errors)
                {
                    sb.AppendFormat("Line {0}: {1}", error.Line, error.Message);
                    sb.AppendLine();
                }

                sb.AppendLine();
                sb.Append("The eventsink will not run until these errors have been corrected and the COM+ application has been restarted.");
                Trace.Error(sb.ToString());


                OnErrorOccured(null, ex);
                return false;
            }
            catch (Exception ex)
            {
                _InitializationFailed = true;
                Trace.Error("The following unhandled occured during the initialization of the event sink: {0}.\r\nThe eventsink will not run until these errors have been corrected and the COM+ application has been restarted\r\n.Stacktace: {1}", ex.Message, ex.StackTrace);
                OnErrorOccured(null, ex);

                return false;
            }

            

            
        }

        protected virtual void OnInitialize()
        {
        }

        protected virtual void OnErrorOccured(EventInfo eventInfo, Exception ex)
        {
        }

        protected static EventInfo GetEventInfo(IExStoreDispEventInfo dispInfo, EventFlags flags, string url)
        {
            return new EventInfo
                       {
                           Connection = new Connection((_Connection) dispInfo.EventConnection, false),
                           Flags = flags,
                           EventIsSynchronous =
                               ((flags &
                                 (EventFlags.EVT_SYNC_ABORTED | EventFlags.EVT_SYNC_BEGIN |
                                  EventFlags.EVT_SYNC_COMMITTED)) != 0),
                           IsFolder = ((flags & EventFlags.EVT_IS_COLLECTION) == EventFlags.EVT_IS_COLLECTION),
                           InternalEventInfo = dispInfo,
                           Item = (Record) dispInfo.EventRecord,
                           SourceUrl = dispInfo.SourceURL,
                           Url = url
                       };
        }

        /// <summary>
        /// Called when the eventsink is registered on a public or private folder.
        /// </summary>
        /// <returns>True, if the registration is allowed. False to prevent the registration from happening.</returns>
        protected virtual bool OnRegisteringEvent(EventInfo info)
        {
            return true;
        }

        static EventSink()
        {
            Trace = new Trace("infinitec.exchange.eventing");
        }
    }
}