﻿using System;
using System.Collections.Generic;
using System.Threading;
using InfiniTec.Diagnostics;
using InfiniTec.Exchange.Eventing;

namespace InfiniTec.Exchange.Eventing
{
    public static class JobProcessor
    {
        private static readonly Queue<Job> _Jobs = new Queue<Job>();
        private static readonly object _JobLock = new object();
        private static readonly AutoResetEvent _JobSignal = new AutoResetEvent(false);
        private static readonly Trace Trace = EventSink.Trace;
        private static Thread _JobExecutionThread;

        public static void Execute(Job job)
        {
            if (job == null) throw new ArgumentNullException("job");

            if (!EventingConfigurationSection.Configuration.JobProcessing.Asynchronous)
            {
                try
                {
                    Trace.Transfer(0, job.Id, "Synchronously executing job {0}", job.Name);
                    job.Execute();
                }
                catch (Exception ex)
                {
                    ProcessException(ex, job);
                }
            }
            else
            {
                lock (_JobLock)
                {
                    if (_JobExecutionThread == null)
                    {
                        StartExecutionThread();
                    }
                    EventSink.Trace.Transfer(0, job.Id, "Job {0} has been queued for asynchronous execution", job.Name);
                    _Jobs.Enqueue(job);
                }
                _JobSignal.Set();
            }
        }

        private static void StartExecutionThread()
        {
            _JobExecutionThread = new Thread(JobProcessingThread) {Name = "Asynchronous job processing thread"};
            _JobExecutionThread.Start();
        }

        private static void ProcessException(Exception ex, Job job)
        {
            Trace.Error("Execution of job {0} failed. Exception: {1}", job.Name, ex);
        }

        private static void JobProcessingThread()
        {
            while (true)
            {
                Diagnostics.Trace.ActivityId = Guid.Empty;
                Job job = null;
                lock (_JobLock)
                {
                    if (_Jobs.Count > 0)
                    {
                        job = _Jobs.Dequeue();
                        Trace.Verbose("Job {0} has been scheduled for asynchronous execution.", job.Name);
                    }
                }
                if (job == null)
                {
                    Trace.Verbose("No jobs are currently scheduled for execution. Waiting indefinetely for new jobs.");
                    _JobSignal.WaitOne();
                    continue;
                }

                try
                {
                    Trace.ActivityId = job.Id;
                    Trace.Verbose("Executing job {0}", job.Name);
                    job.Execute();
                }
                catch (Exception ex)
                {
                    ProcessException(ex, job);
                }
            }
        }
    }
}