﻿using System;
using ADODB;
using InfiniTec.Diagnostics;
using InfiniTec.Exchange.Eventing.Interop;

namespace InfiniTec.Exchange.Eventing
{
    public abstract class SynchronousEventSink : AsynchronousEventSink, IExStoreSyncEvents
    {
        void IExStoreSyncEvents.OnSyncSave(IExStoreEventInfo pEventInfo, string bstrURLItem, EventFlags lFlags)
        {
            Trace.ActivityId = Guid.NewGuid();

            if ((lFlags & EventFlags.EVT_SYNC_BEGIN) == EventFlags.EVT_SYNC_BEGIN)
            {
				Trace.Verbose("OnSyncSave called for item {0} (Begin phase).", bstrURLItem);
                ProcessSynchronousOperation(bstrURLItem, pEventInfo, lFlags, OperationType.Save);
            }
            else if ((lFlags & EventFlags.EVT_SYNC_COMMITTED) == EventFlags.EVT_SYNC_COMMITTED)
			{
				Trace.Verbose("OnSyncSave called for item {0} (Commit phase).", bstrURLItem);
                ProcessCommitOperation(bstrURLItem, pEventInfo, lFlags, OperationType.Save);
            }
            else if ((lFlags & EventFlags.EVT_SYNC_ABORTED) == EventFlags.EVT_SYNC_ABORTED)
            {
				Trace.Verbose("OnSyncSave called for item {0} (Abort phase).", bstrURLItem);
                ProcessAbortOperation(bstrURLItem, pEventInfo, lFlags, OperationType.Save);
            }
        }

        public static void Abort(EventInfo info)
        {
            Abort(1, info);
        }
        public static void Abort(int errorCode, EventInfo info)
        {
            if (info == null) throw new ArgumentNullException("info");

            if ((info.Flags & EventFlags.EVT_SYNC_BEGIN) != EventFlags.EVT_SYNC_BEGIN)
            {
                throw new InvalidOperationException("The Abort method can only be called during the OnItemCreating or OnItemDeleting event.");
            }

            info.InternalEventInfo.AbortChange(errorCode);
        }

        public static void SaveItemSilent(_Record record)
        {
            record.Fields[WellknownProperties.Eventing.IsSilentSaveOperation].Value = true;
            record.Fields.Update();
        }

		public static void MarkAsSilent(EventInfo info)
		{
			info.IsSilentSaveOperation = true;
		}

        void IExStoreSyncEvents.OnSyncDelete(IExStoreEventInfo pEventInfo, string bstrURLItem, EventFlags lFlags)
        {
            Trace.ActivityId = Guid.NewGuid();
            Trace.Verbose("OnSyncDelete called for item {0}.", bstrURLItem);

            if ((lFlags & EventFlags.EVT_SYNC_BEGIN) == EventFlags.EVT_SYNC_BEGIN)
            {
                ProcessSynchronousOperation(bstrURLItem, pEventInfo, lFlags, OperationType.Delete);
            }
            else
            {
                ProcessCommitOperation(bstrURLItem, pEventInfo, lFlags, OperationType.Delete);
            }
        }

        private void ProcessAbortOperation(string url, IExStoreEventInfo eventInfo, EventFlags flags, OperationType operationType)
        {
            string subject = "(unknown)";
            using (EventInfo info = GetEventInfo((IExStoreDispEventInfo) eventInfo, flags, url))
            {
                try
                {
                    if (!Initialize())
                    {
                        Trace.Verbose("The initialization has failed. Aborting.");
                        return;
                    }

                    subject = GetSubject(info);

                    if ((flags & EventFlags.EVT_COPY) == EventFlags.EVT_COPY)
                    {
                        Trace.Verbose("A copy operation on for item {0} was aborted.", info.Url);
                        OnItemCreationAborted(info, CreationMode.Copied);
                    }
                    else if ((flags & EventFlags.EVT_MOVE) == EventFlags.EVT_MOVE && operationType == OperationType.Save)
                    {
                        Trace.Verbose("A move operation on for item {0} was aborted.", info.Url);
                        OnItemCreationAborted(info, CreationMode.Moved);
                    }
                    else if ((flags & EventFlags.EVT_MOVE) == EventFlags.EVT_MOVE &&
                             operationType == OperationType.Delete)
                    {
                        Trace.Verbose("A move to recylclebin operation on for item {0} was aborted.", info.Url);
                        OnItemDeletionAborted(info, DeletionMode.MovedToRecycleBin);
                    }

                    else if ((flags & EventFlags.EVT_HARDDELETE) == EventFlags.EVT_HARDDELETE)
                    {
                        Trace.Verbose("A hard-delete operation on for item {0} was aborted.", info.Url);
                        OnItemDeletionAborted(info, DeletionMode.Hard);
                    }
                    else if ((flags & EventFlags.EVT_SOFTDELETE) == EventFlags.EVT_SOFTDELETE)
                    {
                        Trace.Verbose("A soft-delete operation on for item {0} was aborted.", info.Url);
                        OnItemDeletionAborted(info, DeletionMode.Soft);
                    }
                    else if ((flags & EventFlags.EVT_IS_DELIVERED) == EventFlags.EVT_IS_DELIVERED)
                    {
                        Trace.Verbose("A deliver operation on for item {0} was aborted.", info.Url);
                        OnItemCreationAborted(info, CreationMode.Delivered);
                    }
                    else if ((flags & EventFlags.EVT_NEW_ITEM) == EventFlags.EVT_NEW_ITEM)
                    {
                        Trace.Verbose("A new item operation on for item {0} was aborted.", info.Url);
                        OnItemCreationAborted(info, CreationMode.NewItem);
                    }
                    else if ((flags & EventFlags.EVT_REPLICATED_ITEM) == EventFlags.EVT_REPLICATED_ITEM)
                    {
                        Trace.Verbose("A replicate operation for item {0} was aborted.", info.Url);
                        OnItemCreationAborted(info, CreationMode.Replicated);
                    }
                    else
                    {
                        Trace.Verbose("An update operation for item {0} was aborted.", info.Url);
                        OnItemUpdateAborted(info);
                    }
                }
                catch (Exception ex)
                {
                    Trace.Error("An unhandled exception occured while processing item at {0} (Subject {1}): {2}",
                                       info.Url, subject, ex);

                    OnErrorOccured(info, ex);
                }
            }
        }

        protected virtual void OnItemUpdateAborted(EventInfo info)
        {
        }

        protected virtual void OnItemDeletionAborted(EventInfo info, DeletionMode deletionMode)
        {
        }

        protected virtual void OnItemCreationAborted(EventInfo info, CreationMode creationMode)
        {
        }


        private void ProcessSynchronousOperation(string url, IExStoreEventInfo eventInfo, EventFlags flags, OperationType operationType)
        {
            var statebag = new StateBag();

            using (EventInfo info = GetEventInfo((IExStoreDispEventInfo) eventInfo, flags, url))
            {
                try
                {
                    if (!Initialize())
                    {
                        Trace.Verbose("The initialization has failed. Aborting.");
                        return;
                    }

                    var value = info.Item.Fields[WellknownProperties.Eventing.IsSilentSaveOperation].Value as bool?;
                    if (value.GetValueOrDefault())
                    {
                        Trace.Verbose("The save operation has been marked as silent. Aborting event sink.");
                        statebag.IsSilentSaveOperation = true;
                        info.Item.Fields[WellknownProperties.Eventing.IsSilentSaveOperation].Value = null;
                        info.Item.Fields.Update();

                        OnItemSilentlySaving(info);

                        return;
                    }

                    if ((flags & EventFlags.EVT_COPY) == EventFlags.EVT_COPY)
                    {
                        Trace.Verbose("The item {0} is being copied here from {1}", info.Url, info.SourceUrl);
                        OnItemCreating(info, CreationMode.Copied);
                    }
                    else if ((flags & EventFlags.EVT_MOVE) == EventFlags.EVT_MOVE && operationType == OperationType.Save)
                    {
                        Trace.Verbose("The item {0} is being moved here from {1}", info.Url, info.SourceUrl);
                        OnItemCreating(info, CreationMode.Moved);
                    }
                    else if ((flags & EventFlags.EVT_MOVE) == EventFlags.EVT_MOVE && operationType == OperationType.Delete)
                    {
                        Trace.Verbose("The item {0} is moved to the recycle-bin", info.Url);
                        OnItemDeleting(info, DeletionMode.MovedToRecycleBin);
                    }

                    else if ((flags & EventFlags.EVT_HARDDELETE) == EventFlags.EVT_HARDDELETE)
                    {
                        Trace.Verbose("The item {0} is being hard-deleted from the store", info.Url);
                        OnItemDeleting(info, DeletionMode.Hard);
                    }
                    else if ((flags & EventFlags.EVT_SOFTDELETE) == EventFlags.EVT_SOFTDELETE)
                    {
                        Trace.Verbose("The item {0} is being soft-deleted from the store", info.Url);
                        OnItemDeleting(info, DeletionMode.Soft);
                    }
                    else if ((flags & EventFlags.EVT_IS_DELIVERED) == EventFlags.EVT_IS_DELIVERED)
                    {
                        Trace.Verbose("A new item is being delivered to {0}", info.Url);
                        OnItemCreating(info, CreationMode.Delivered);
                    }
                    else if ((flags & EventFlags.EVT_NEW_ITEM) == EventFlags.EVT_NEW_ITEM)
                    {
                        Trace.Verbose("A new item is being created at {0}", info.Url);
                        OnItemCreating(info, CreationMode.NewItem);
                    }
                    else if ((flags & EventFlags.EVT_REPLICATED_ITEM) == EventFlags.EVT_REPLICATED_ITEM)
                    {
                        Trace.Verbose("A new item is being replicated to {0}", info.Url);
                        OnItemCreating(info, CreationMode.Replicated);
                    }
                    else if ((flags & EventFlags.EVT_REPLICATED_ITEM) == EventFlags.EVT_REPLICATED_ITEM)
                    {
                        Trace.Verbose("The existing item {0} is being updated.", info.Url);
                        OnItemUpdating(info);
                    }
                    else
                    {
                        Trace.Verbose("The item {0} is being updated.", info.Url);
                        OnItemUpdating(info);
                    }
                }
            
                catch (Exception ex)
                {
                    Trace.Error("An unhandled exception occured while processing item at {0}: {1}", info.Url, ex);

                    OnErrorOccured(info, ex);
                }
                finally
                {
					if (info.IsSilentSaveOperation)
					{
						Trace.Verbose("The save operation has been marked as silent.", info.Url);
						statebag.IsSilentSaveOperation = true;
					}
                    statebag.UserState = info.UserState;
                    info.InternalEventInfo.Data = statebag.SerializeToUnmanagedMememory();
                }
            }
        }

        protected virtual void OnItemSilentlySaving(EventInfo info)
        {
            
        }

        protected virtual void OnItemUpdating(EventInfo info)
        {
        }

        /// <summary>
        /// Called when an item or folder is deleted. The <see cref="EventInfo.Item"/> allows read-write access to the newly created item.
        /// </summary>
        /// <param name="info">A <see cref="EventInfo"/> instance containing detailed informations about the item being deleted.</param>
        /// <param name="deletionMode">A value of the <see cref="CreationMode"/> specifying whether the has been deleted or moved to the recycle bin.</param>
        /// <remarks>
        /// <para>Note to implementors: The Exchange store is blocked during the execution of this method. This method must be executed as fast as possible and no
        /// blocking calls should be made from this thread. Long running operations should be executed on a different thread.</para>
        /// <para>Thread safety: This method is thread safe.</para>
        /// </remarks>
        protected virtual void OnItemDeleting(EventInfo info, DeletionMode deletionMode)
        {
        }


        /// <summary>
        /// Called when a new item is created in a folder. The <see cref="EventInfo.Item"/> allows
        /// read-write access to the newly created item.
        /// </summary>
        /// <param name="info">A <see cref="EventInfo"/> instance containing detailed informations about the newly created item.</param>
        /// <param name="creationMode">A value of the <see cref="CreationMode"/> specifying whether the item is new or was copied/moved to the current
        /// location.</param>
        /// <remarks>
        /// <para>Note to implementors: The Exchange store is blocked during the execution of this method. This method must be executed as fast as possible and no
        /// blocking calls should be made from this thread. Long running operations should be executed on a different thread.</para>
        /// <para>Thread safety: This method is thread safe.</para>
        /// </remarks>
        protected virtual void OnItemCreating(EventInfo info, CreationMode creationMode)
        {
        }
    }
}